/*
 * notch.c - routines for producing a drive notch report.  Generally works 
 *           only for later Seagate drives (eg. ST5150N).
 *
 * Can report the Notch and Partition Page, to the extent the drive
 * supports it (many don't very well) by setting the Active Notch and
 * iterating.
 * 
 * Does a SEND/RECEIVE DIAGNOSTIC with the Translate Address page for
 * every sector on the drive.
 *
 * Rod Van Meter (rdv@alumni.caltech.edu, http://www.alumni.caltech.edu/~rdv/)
 *
 * Author: Rod Van Meter
 *         University of Southern California, Information Sciences Institute
 *         rdv@isi.edu
 */

#include <stdio.h>
#include "uscsi.h"

#ifdef NOTCH_REPORT

#ifndef lint
static char rcsid[] = "$Id: notch.c,v 1.9 1999/04/19 23:41:07 jdd Exp $";
#endif

#ifdef FILENAME_MAX     /* defined in stdio.h only in SYSV systems */
#define SYSV
#include <string.h>
#else
#include <strings.h>
#endif

#include <sys/param.h>
#include <values.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "utils.h"

extern int d;

#ifdef SYSV

#include <stdlib.h>

/* silly workaround for a bug in gcc <<2.6.2 for Solaris 2.4 */
#include <sys/byteorder.h>
#if !defined(_BIT_FIELDS_LTOH) && !defined(_BIT_FIELDS_HTOL)
#ifdef __i386__
#define _BIT_FIELDS_LTOH
#else /* !__i386__ */
#define _BIT_FIELDS_HTOL
#endif /* !__i386__ */
#endif /* !defined(_BIT... */

#include <sys/scsi/generic/mode.h>
#include <sys/scsi/generic/commands.h>
#include <sys/scsi/impl/types.h>
#include <sys/scsi/impl/uscsi.h>
#define bzero(b,l) memset(b,0,l)

#else /* SYSV */

#include <scsi/impl/types.h>
#include <scsi/impl/uscsi.h>

#if !defined(_BIT_FIELDS_LTOH) && !defined(_BIT_FIELDS_HTOL)
#if !defined(vax) && !defined(i386) 
/* SunOS 4.x assumes that if it's not intel or vax, it's big-endian. */
#define _BIT_FIELDS_HTOL
#else /* !defined(vax)... */
#define _BIT_FIELDS_LTOH
#endif /* !defined(vax)... */
#endif /* !defined(_BIT... */

#endif /* SYSV */

#define SCSIBUFLEN 8000

struct mode_notch {
	struct	mode_page mode_page;	/* common mode page header */
#if defined(_BIT_FIELDS_LTOH)
	u_char  reserved : 6,
	        lpn : 1,	/* logical or physical notch? */
                nd : 1;		/* notched drive? */
#elif defined(_BIT_FIELDS_HTOL)
	u_char	nd : 1,		/* notched drive? */
		lpn : 1,	/* logical or physical notch? */
		reserved : 6;
#else
error: One of _BIT_FIELDS_LTOH or _BIT_FIELDS_HTOL must be defined
#endif  /* _BIT_FIELDS_LTOH */
	u_char  reserved1;
	u_char  maxnotch_ub;
	u_char  maxnotch_lb;
	u_char  active_notch_ub;
	u_char  active_notch_lb;
	u_int   start;
	u_int   end;
	u_int   pages_notched_ul;
	u_int   pages_notched_ll;
};

#define NOTCH_PAGE_LENGTH 0x16
#define MODE_SELECT_PAGE_FORMAT 0x10

struct diag_addr_trans {
    /* struct diag_page diag_page;	/ * common header */
    u_char	page_code;
    u_char	reserved;
    u_short	length;
    u_char	supplied;	/* actually 3-bit field */
    u_char	translate;
    u_char	addr[8];
};

#define TRANSLATE_PAGE_LENGTH	0x0a
#define DAD_TRANSLATE		0x40
#define TRANS_BLOCK_FORMAT	0x00
#define TRANS_PHYS_FORMAT	0x05
#define SEND_DIAG_PAGE_FORMAT	0x10 /* cdb byte 1 */

static int scsi_set_active_notch(fd,notch)
	int fd;
	int notch;
{
    struct uscsi_cmd cmd;
    union scsi_cdb cdb;
    struct mode_notch mnotch;
    char buff[SCSIBUFLEN];
    char *buffp;
    int i;
    
    bzero(&cdb, sizeof(cdb));
    bzero(&mnotch, sizeof(mnotch));
    bzero(buff,SCSIBUFLEN);
    mnotch.mode_page.length = NOTCH_PAGE_LENGTH;
    mnotch.mode_page.code = DAD_MODE_NOTCHPART;
    mnotch.active_notch_ub = notch >> 8;
    mnotch.active_notch_lb = notch & 0xff;
    cdb.scc_cmd = SCMD_MODE_SELECT;
    cdb.g0_count0 = sizeof(mnotch)+MODE_HEADER_LENGTH;
    cdb.cdb_un.tag = MODE_SELECT_PAGE_FORMAT;
    cmd.uscsi_cdb=(caddr_t)&cdb;
    cmd.uscsi_cdblen=6; /* SCSI Group 0 cmd */
    cmd.uscsi_bufaddr=(caddr_t)buff;
    cmd.uscsi_buflen=sizeof(mnotch)+MODE_HEADER_LENGTH;
    cmd.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE;

    buff[0] = sizeof(mnotch)+MODE_HEADER_LENGTH - 1;
    buffp = MODE_PAGE_ADDR((struct mode_header *)buff, char);
    /* don't need both, since we already know */
    /* memcpy(buffp, (char *)&mnotch, sizeof(struct mode_page)); */
    memcpy(buffp, (char *)&mnotch, mnotch.mode_page.length);

    /* write the notch page */
    /* cdb.g0_addr1 = DAD_MODE_NOTCHPART; */

#ifdef ZONE_DEBUG
    printf("cdb: ");
    for (i = 0 ; i < 6 ; i++ )
	printf("0x%02x ",((char *)&cdb)[i]);
    printf("\ndata: ");
    for (i = 0 ; i < NOTCH_PAGE_LENGTH+MODE_HEADER_LENGTH ; i++ )
	printf("0x%02x ",buff[i]);
    printf("\n");
#endif

    if(-1==ioctl(fd, USCSICMD, &cmd)) return(-1);
    
    return(0);
}

static int scsi_read_active_notch(fd,maxnotchp,notchp,startp,endp,
		pagesup,pageslp)
	int fd;
	int *maxnotchp,*notchp,*startp,*endp,*pagesup,*pageslp;
{
    struct uscsi_cmd cmd;
    union scsi_cdb cdb;
    struct mode_notch mnotch;
    char buff[SCSIBUFLEN];
    char *buffp;
    
    bzero(&cdb, sizeof(cdb));
    bzero(&mnotch, sizeof(mnotch));

    cdb.scc_cmd = SCMD_MODE_SENSE;
    cdb.g0_count0 = 255;
    cmd.uscsi_cdb=(caddr_t)&cdb;
    cmd.uscsi_cdblen=6; /* SCSI Group 0 cmd */
    cmd.uscsi_bufaddr=(caddr_t)buff;
    cmd.uscsi_buflen=SCSIBUFLEN;
    cmd.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ;

    /* read the notch page */
    cdb.g0_addr1 = DAD_MODE_NOTCHPART;
    if(-1==ioctl(fd, USCSICMD, &cmd)) return(-1);
    
    buffp = MODE_PAGE_ADDR((struct mode_header *)buff, char);
    memcpy((char *)&mnotch, buffp, sizeof(struct mode_page));
    memcpy((char *)&mnotch, buffp, mnotch.mode_page.length);

    *maxnotchp = (mnotch.maxnotch_ub << 8) + mnotch.maxnotch_lb;
    *notchp = (mnotch.active_notch_ub << 8) + mnotch.active_notch_lb;
    *startp = mnotch.start;
    *endp = mnotch.end;
    *pagesup = mnotch.pages_notched_ul;
    *pageslp = mnotch.pages_notched_ll;

    return(0);
}

static int scsi_translate_address_receive(fd,diskdevice,highp,lowp)
	int fd;
	char *diskdevice;
	int *highp,*lowp;
{
    struct uscsi_cmd cmd;
    union scsi_cdb cdb;
    struct diag_addr_trans dat;
    char buff[SCSIBUFLEN];
    int i;
    
    bzero(&cdb, sizeof(cdb));
    bzero(&dat, sizeof(dat));
    bzero(buff, sizeof(dat));

    cdb.scc_cmd = SCMD_GDIAG;
    cdb.g0_count0 = sizeof(dat);

    cmd.uscsi_cdb=(caddr_t)&cdb;
    cmd.uscsi_cdblen=6; /* SCSI Group 0 cmd */
    cmd.uscsi_bufaddr=(caddr_t)buff;
    cmd.uscsi_buflen = sizeof(dat);
    cmd.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE|USCSI_READ;

#ifdef ZONE_DEBUG
    printf("cdb: ");
    for (i = 0 ; i < 6 ; i++ )
	printf("0x%02x ",((char *)&cdb)[i]);
    printf("\ndata: ");
    for (i = 0 ; i < sizeof(dat) ; i++ )
	printf("0x%02x ",buff[i]);
    printf("\n");
#endif

    if(-1==ioctl(fd, USCSICMD, &cmd)) return(-1);
    
#ifdef ZONE_DEBUG
    printf("buff: ");
    for (i = 0 ; i < sizeof(dat) ; i++ )
	printf("0x%02x ",buff[i]);
    printf("\n");
#endif

    /*
    memcpy((char *)&dat, buff, sizeof(struct mode_page));
    memcpy((char *)&dat, buff, dat.length);
     */
    memcpy((char *)&dat, buff, sizeof(dat));

    *highp = (dat.addr[0] << 24) + (dat.addr[1] << 16) + (dat.addr[2] << 8)
	+ dat.addr[3];
    *lowp = (dat.addr[4] << 24) + (dat.addr[5] << 16) + (dat.addr[6] << 8)
	+ dat.addr[7];
    return(0);
}

static int scsi_translate_address_send(fd,diskdevice,block)
	int fd;
	char *diskdevice;
	int block;
{
    struct uscsi_cmd cmd;
    union scsi_cdb cdb;
    struct diag_addr_trans dat;
    char buff[SCSIBUFLEN];
    int i;
    
    bzero(&cdb, sizeof(cdb));
    bzero(&dat, sizeof(dat));

    cdb.scc_cmd = SCMD_SDIAG;

    cdb.cdb_un.tag = SEND_DIAG_PAGE_FORMAT;
    cdb.g0_count0 = sizeof(dat);

    /* SELF TEST
    cdb.cdb_un.tag = 4;
    cdb.g0_count0 = 0;
     */

    dat.length = TRANSLATE_PAGE_LENGTH;
    dat.page_code = DAD_TRANSLATE;
    dat.addr[0] = block >> 24;
    dat.addr[1] = (block >> 16) & 0xff;
    dat.addr[2] = (block >> 8) & 0xff;
    dat.addr[3] = block & 0xff;
    dat.supplied = TRANS_BLOCK_FORMAT;
    dat.translate = TRANS_PHYS_FORMAT;

    cmd.uscsi_cdb=(caddr_t)&cdb;
    cmd.uscsi_cdblen=6; /* SCSI Group 0 cmd */
    cmd.uscsi_bufaddr=(caddr_t)buff;
    cmd.uscsi_buflen = sizeof(dat);
    cmd.uscsi_flags=USCSI_DIAGNOSE|USCSI_ISOLATE;

    memcpy(buff, (char *)&dat, sizeof(dat));

#ifdef ZONE_DEBUG
    printf("cdb: ");
    for (i = 0 ; i < 6 ; i++ )
	printf("0x%02x ",((char *)&cdb)[i]);
    printf("\ndata: ");
    for (i = 0 ; i < sizeof(dat) ; i++ )
	printf("0x%02x ",buff[i]);
    printf("\n");
#endif

    return ioctl(fd, USCSICMD, &cmd);
}

static void scsi_translate_address(fd,diskdevice,block,highp,lowp)
	int fd;
	char *diskdevice;
	int block,*highp,*lowp;
{
    if (0>scsi_translate_address_send(fd,diskdevice,block)) {
	perror(diskdevice);
	Error("SCSI send diag failed, giving up.");
	return;
    }
    if (0>scsi_translate_address_receive(fd,diskdevice,highp,lowp)) {
	perror(diskdevice);
	Error("SCSI rcv diag failed, giving up.");
	return;
    }
}

static void notches(fd,diskdevice,capacity)
	int fd;
	char *diskdevice;
	unsigned long capacity;
{
    int i;
    int maxnot,maxnotches,notch,start,end,pagesu,pagesl,
        high_ba = 0,low_ba = 0,last_hba = -2,last_lba = -2;

    if(0>scsi_read_active_notch(fd, &maxnotches, &notch,
				&start, &end, &pagesu, &pagesl)) {
	perror(diskdevice);
	Error("SCSI mode sense failed. Drive does not support notch queries.");
    }
    printf("\nNotch Page 0: maxnotches: %d notch: %d start: %d end: %d pages: 0x%x 0x%x\n",
	   maxnotches,notch,start,end,pagesu,pagesl);
	printf("i\tmaxnot\tnotch\tstart\t\tend\t\tpagesU\tpagesL\n");

    maxnot = maxnotches;
    for ( i = 1 ; i <= maxnot ; i++ ) {
	if(0>scsi_set_active_notch(fd,i)) {
	    printf("SCSI mode sense failed at notch %d\n",i); fflush(stdout);
	    perror(diskdevice);
	    /* Error("SCSI mode select failed, giving up."); */
	    break;
	}
	if(0>scsi_read_active_notch(fd, &maxnotches, &notch,
				    &start, &end, &pagesu, &pagesl)) {
	    printf("SCSI mode sense failed at notch %d\n",i); fflush(stdout);
	    perror(diskdevice);
	    /* Error("SCSI mode sense failed, giving up."); */
	    /* printf("SCSI mode sense failed, giving up.\n"); */
	    break;
	}
	printf("%d\t%d\t%d\t0x%08x\t0x%08x\t0x%x\t0x%x\n",
	       i,maxnotches,notch,start,end,pagesu,pagesl);
    }
    if(0>scsi_set_active_notch(fd,0)) {
	printf("SCSI mode select failed when resetting active notch\n"); fflush(stdout);
	perror(diskdevice);
	/* Error("SCSI mode select failed, giving up."); */
    }
    printf("\nblock\tcyl\thead\tsector\n");
    for ( i = 0 ; i < capacity ; i++ ) {
	scsi_translate_address(fd,diskdevice,i,&high_ba,&low_ba);
	if (high_ba != last_hba || low_ba != last_lba +1) {
	    if (i)
		printf("%d\t0x%04x\t0x%02x\t0x%02x\n",
		       i-1,last_hba >> 8, last_hba & 0xff, last_lba);
	    printf("%d\t0x%04x\t0x%02x\t0x%02x\n",
		   i, high_ba >> 8, high_ba & 0xff, low_ba);
	}
	last_hba = high_ba; last_lba = low_ba;
    }
    return;
}

void report_notches(device)
char *device;
{
	unsigned long capacity;
	struct stat sb;
	int fd;

	if(stat(device, &sb)<0){
		perror(device);
		Error("Cannot access device, giving up.");
	}
	if(!S_ISCHR(sb.st_mode)){
		Error("%s must be a raw disk device.", device);
	}
	if(0>(fd=open(device, O_RDONLY|O_NDELAY))){
		perror(device);
		Error("Cannot open device, giving up.");
	}
	/* get number of blocks available */
	if(0>scsi_capacity(fd, &capacity)){
		perror(device);
		Error("SCSI READ_CAPACITY failed, giving up.");
	}
	notches(fd, device, capacity);
}

#endif /* NOTCH_REPORT */
