/*
 * scsiinfo.c - main src file.
 *
 * Author: John DiMarco, University of Toronto, CSLab <jdd@cs.toronto.edu>
 */

#ifndef lint
static char rcsid[] = "$Id: scsiinfo.c,v 3.27 2000/03/30 22:06:36 jdd Exp $";
#endif

#include "patchlevel.h"
#include "uscsi.h"
#include "format.h"
#include <stdio.h>

#ifdef FILENAME_MAX     /* defined in stdio.h only for SunOS >=5.x */
#define SYSV
#endif

#include <sys/param.h>
#include <values.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <nlist.h>
#include <kvm.h>
#include <fcntl.h>
#include <sys/utsname.h>
#include "utils.h"
#include "param.h"

/* os-independent globals */
char *progname;		/* program name */
int d;			/* debugging */
int parameter = 0; 	/* for param.h */
kvm_t *kd; 		/* kernel file descriptor */
char *osrelease;	/* for OS revision checking */
u_int devicecount = 0;	/* number of SCSI devices found */

/* include OS-dependent stuff */
#define GLOBALS /* ensure os-dependent globals are defined */
#ifdef SYSV
#include "sunos5.h"
#else 
#include "sunos4.h"
#endif

#ifdef ESP_SUPPORT
extern int process_esp();
#endif /* ESP_SUPPORT */

#ifdef ISP_SUPPORT
extern int process_isp();
#endif /* ISP_SUPPORT */

#ifdef FAS_SUPPORT
extern int process_fas();
#endif /* FAS_SUPPORT */

#ifdef PTISP_SUPPORT
extern int process_ptisp();
#endif /* PTISP_SUPPORT */

#ifdef GLM_SUPPORT
extern int process_glm();
#endif /* GLM_SUPPORT */

#ifdef PLN_SUPPORT
extern int process_pln();
#endif /* PLN_SUPPORT */

#ifdef NOTCH_REPORT
#include "notch.h"
static int notchreport=0; /* Produce a notch report */
#endif /* NOTCH_REPORT */

static int noisycount=0; /* number of noisy devices */

static void usage()
{
	fprintf(stderr, "\
Usage: %s [-V][-o][-r device][-f diskdev][-F diskdev]%s\n\
       [-c][-n][-s][-p][-q][-O][-K kernelimage][-C coreimage]\n",  progname, 
#ifdef NOTCH_REPORT
	"[-N diskdev]");
#else /* !NOTCH_REPORT */
	"");
#endif /* !NOTCH_REPORT */
	exit(-1);
}

int main(argc, argv)
int argc;
char *argv[];
{
	int c;
	int errflg = 0;
	extern char *optarg;
	static struct utsname un;
	static char *device=NULL;
#ifdef NOTCH_REPORT
	static char *notchdevice;
#endif /* NOTCHREPORT */
	char *kernelimage=NULL, *coreimage=NULL;

	/* synch messages are on by default, unless -n flag */
	set_param(P_SYNCH);

	progname = argv[0];
	while ((c = getopt(argc, argv, "VodcsnOpqr:f:F:N:K:C:")) != EOF)
		switch (c) {
		case 'V':
			printf("%s\n", patchlevel);
			printf("Compiled for SunOS %s:", 
				SUNOS_VERSION[0]?SUNOS_VERSION:"?.?");
#ifdef ESP_SUPPORT
			printf(" ESP");
#endif /* ESP_SUPPORT */
#ifdef PTISP_SUPPORT
			printf(" PTISP");
#endif /* PTISP_SUPPORT */

#ifdef SYSV
	/* adapters below are only supported for SunOS 5.x */
#ifdef ISP_SUPPORT
			printf(" ISP");
#endif /* ISP_SUPPORT */
#ifdef FAS_SUPPORT
			printf(" FAS");
#endif /* FAS_SUPPORT */
#ifdef GLM_SUPPORT
			printf(" GLM");
#endif /* GLM_SUPPORT */
#ifdef PLN_SUPPORT
			printf(" PLN");
#endif /* PLN_SUPPORT */
#ifdef __sparcv9
			printf(" 64BIT");
#endif /* __sparcv9 */
#endif /* SYSV */

#ifdef NOTCH_REPORT
			printf(" NOTCH");
#endif /* NOTCH_REPORT */
			printf(" FORMAT");
			printf("\n");
			exit(0);
			break;
		case 'd':
			++d;
			break;
		case 'c': 
			/* output controller information */
			set_param(P_CONTROLLER);
			break;
		case 's':
			/* Use SunOS4.x boot-style "now Synchronous" format */
			set_param(P_SUNOS4STYLE);
			break;
		case 'p':
			/* probe probe-able devices */
			set_param(P_PROBE);
			break;
		case 'n':
			/* disable synch messages */
			clear_param(P_SYNCH);
			break;
		case 'q':
			/* quiet mode */
			set_param(P_QUIET);
			break;
		case 'o':
			/* use manufacturer's original settings for fmt info */
			set_param(P_ORIGINAL);
			break;
		case 'F':
			/* output format.dat */
			set_param(P_FORMATDAT);
			/* fall thru */
		case 'f': 
		case 'r': 
			/* output (format) report */
			set_param(P_REPORT);
			device=optarg;
			break;
#ifdef NOTCH_REPORT
		case 'N': 
			/* produce disk notch report */
			notchreport++;
			notchdevice=optarg;
			break;
#endif /* NOTCH_REPORT */
		case 'K': 
			/* kernel image */
			kernelimage=optarg;
			break;
		case 'C': 
			/* core image */
			coreimage=optarg;
			break;
		case 'O':
			/* check if something has device open before probing */
			set_param(P_OPENCHECK);
			break;
		default:
			/* parameter error */
			errflg++;
			break;
		}
	if (errflg) {
		usage();
	}

	/* get OS version and machinetype */
	(void)uname(&un);
	osrelease=s(un.release);


	/* generate format information if specified */
	if(NULL!=device && !STREQ(device, "all")){
		if(param_is_set(P_REPORT)){
			report(param_is_set(P_FORMATDAT), osrelease, device, 
					param_is_set(P_ORIGINAL));
			exit(0);
		} 
	}
#ifdef NOTCH_REPORT
	if(notchreport){
		report_notches(notchdevice);
		exit(0);
	}
#endif /* NOTCH_REPORT */

#ifndef SKIP_OSCHECK
	/* quit if we have an OS or kernel architecture mismatch */
	if(STREQ("", SUNOS_VERSION)){
		Warning(
		"WARNING: SUNOS_VERSION was not specified at compile time.");
	} else {
		if(!STREQ(SUNOS_VERSION, un.release)){
			Warning(
		  "WARNING: compiled for SunOS %s; this is SunOS %s. Results may be incorrect.",
		      SUNOS_VERSION, un.release);
		}
	}
	if((0!=strncmp(un.machine,"sun4",4)) || strlen(un.machine)<(u_int)5){
		Error( "%s kernel architecture not supported.", un.machine);
	}
#endif /* SKIP_OSCHECK */

	/* open the kernel */
        if(NULL==(kd=kvm_open(kernelimage,coreimage,NULL,O_RDONLY,progname))){
		Error("Cannot read kernel, giving up.");
        }
        if(0>kvm_nlist(kd, nl)){
                Error("kernel symbols read unsuccessful");
        }

	/* Now process the defined controllers */
#ifdef ESP_SUPPORT
	noisycount += process_esp();
#endif /* ESP_SUPPORT */

#ifdef ISP_SUPPORT
	noisycount += process_isp();
#endif /* ISP_SUPPORT */

#ifdef FAS_SUPPORT
	noisycount += process_fas();
#endif /* FAS_SUPPORT */

#ifdef PTISP_SUPPORT
	noisycount += process_ptisp();
#endif /* PTISP_SUPPORT */

#ifdef GLM_SUPPORT
	noisycount += process_glm();
#endif /* GLM_SUPPORT */

#ifdef PLN_SUPPORT
	noisycount += process_pln();
#endif /* PLN_SUPPORT */
	return(noisycount);
}
