/*
 * utils.c: miscellaneous utilities
 *
 * Author: John DiMarco, University of Toronto, CDF
 *         jdd@cdf.toronto.edu
 */

#ifndef lint
static char rcsid[] = "$Id: utils.c,v 1.6 1995/03/28 18:46:39 jdd Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef FILENAME_MAX
#include <string.h>
#else
#include <strings.h>
#endif
#include <varargs.h>
#include "utils.h"

#ifdef lint
#undef va_arg
#define va_arg(x,y) (y)NULL
#endif

extern char *progname;
extern int d;

/*
 * Error(): behaves like fprintf(stderr, ...) followed by exit(-1), except
 *          that the 'programname: ' preceeds the print, and a newline
 *	    follows it.
 */
/*VARARGS*/
void Error(va_alist)
va_dcl
{
	va_list args;
	char *format;

	va_start(args);
	format = va_arg(args, char *);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, format, args);
	fprintf(stderr, "\n");
	va_end(args);
	exit(-1);
}

/*
 * dfprintf(): behaves like fprintf, except the first argument must be a
 *            debugging level. The message will only be printed if "d"
 *            is equal to or greater than the debugging level.
 */
/*VARARGS*/
void dfprintf(va_alist)
va_dcl
{
	va_list args;
	int debugLevel;
	FILE *stream;
	char *format;
	
	va_start(args);
	debugLevel = va_arg(args, int);
	stream = va_arg(args, FILE *);
	format = va_arg(args, char *);
	if(d >= debugLevel){
		vfprintf(stream, format, args);
	}
	va_end(args);
}
	
/*
 * Warning(): behaves like Error, except returns rather than exits.
 */
/*VARARGS*/
void Warning(va_alist)
va_dcl
{
	va_list args;
	char *format;

	va_start(args);
	format = va_arg(args, char *);
    fprintf(stderr, "%s: ", progname);
    vfprintf(stderr, format, args);
    fprintf(stderr, "\n");
    va_end(args);
}

FILE *efopen(file, mode)
char *file, *mode;
{
	FILE *fp;
	if (NULL!=(fp=fopen(file,mode)))
		return(fp);
	Error("can't open file \"%s\" mode \"%s\"", file, mode);
	/*NOTREACHED*/
}

void efclose(f)
FILE *f;
{
	if(EOF==fclose(f))
		Error("can't close file");
	/*NOTREACHED*/
}

/*
 * mylib_malloc(): Checks if it gets a NULL pointer, calls Error if so.
 */
char *mylib_malloc(size, file, line)
unsigned size;
char *file;
int line;
{
	char *result;

	result = malloc(size);
	if(NULL==result){
		Error("Out of memory at line %d in \"%s\".", line, file);
	}
	return(result);
}

/*
 * mylib_realloc(): Checks if it gets a NULL pointer, calls Error if so.
 */
char *mylib_realloc(ptr, size, file, line)
char *ptr, *file;
unsigned size;
int line;
{
	char *result;

	result = realloc(ptr, size);
	if(NULL==result){
		Error("Out of memory at line %d in \"%s\".", line, file);
	}
	return(result);
}

/*
 * mylib_scopy(): Takes a string and creates a new physical copy of it.
 */
char *mylib_scopy(string, file, line)
char *string, *file;
int line;
{
	char *result;

	result = malloc((unsigned)strlen(string)+1);

	if(NULL==result){
		Error("Out of memory at line %d in \"%s\".", line, file);
	}
	(void)strcpy(result, string);
	return(result);
}

/*
 * mylib_srcopy(): Reallocs first string to make room for second, copies it.
 */
char *mylib_srcopy(s1, s2, file, line)
char *s1, *s2, *file;
int line;
{
	s1=mylib_realloc(s1, (unsigned)strlen(s2)+1, file, line);
	(void)strcpy(s1, s2);
	return(s1);
}

/*
 * cat(): Take a list of strings, followed by NULL, return their concatenation
 *        in malloc'ed space.
 */
/*VARARGS*/
char *cat(va_alist)
va_dcl
{
	va_list args;
	unsigned length=1;
	char *str, *newstr;

	/* get length */
	va_start(args);
	loop{
		str = va_arg(args, char *);
		if(NULL!=str){
			 length+=strlen(str);
		} else {
			break;
		}
	}
	va_end(args);

	newstr=malloc(length);
	if(NULL==newstr) Error("Out of memory in cat()");
	
	newstr[0]=(char)0;

	/* create string */
	va_start(args);
	loop{
		str = va_arg(args, char *);
		if(NULL!=str) {
			(void)strcat(newstr, str);
		} else {
			break;
		}
	}
 va_end(args);
#ifdef lint
	args=args; /* make lint shut up about "args set but not used" */
#endif
	return(newstr);
}
